"""Inspired by event.py from https://github.com/jeffthibault/python-nostr.git."""
import binascii
import datetime
import json
import time
from dataclasses import dataclass
from enum import IntEnum
from hashlib import sha256
from typing import List, Optional

from .bech32 import bech32_encode
from .key import PrivateKey, PublicKey
from .message_type import ClientMessageType


class EventKind(IntEnum):
    SET_METADATA = 0
    TEXT_NOTE = 1
    RECOMMEND_RELAY = 2
    CONTACTS = 3
    ENCRYPTED_DIRECT_MESSAGE = 4
    DELETE = 5
    REACTION = 7
    CHANNEL_CREATE = 40
    CHANNEL_META = 41
    CHANNEL_MESSAGE = 42
    CHANNEL_HIDE = 43
    CHANNEL_MUTE = 44
    RELAY_LIST_METADATA = 10002


@dataclass
class Event:
    """Event class.

    :param content: content string
    :param pukey: public key in hex form
    :param created_at: event creation date
    :param kind: event kind
    :param tags: list of list of strings
    :param id: event id, will be computed
    :param sig: signature, will be created after signing with a private key
    """

    content: Optional[str] = None
    pubkey: Optional[str] = None
    created_at: Optional[int] = None
    kind: Optional[int] = EventKind.TEXT_NOTE
    tags: List[List[str]] = None
    id: Optional[str] = None
    sig: Optional[str] = None

    def __post_init__(self):
        if self.content is not None and not isinstance(self.content, str):
            # DMs initialize content to None but all other kinds should pass in a str
            raise TypeError("Argument 'content' must be of type str")
        elif (
            self.content is not None
            and "?iv" not in self.content
            and self.kind == EventKind.ENCRYPTED_DIRECT_MESSAGE
        ):
            raise Exception(
                "Encrypted DMs cannot use the `content` field; use encrypt_dm()"
                "for storing an encrypted content."
            )
        if self.created_at is None:
            self.created_at = int(time.time())

        if self.tags is None:
            self.tags = []

        if self.id is None:
            self.compute_id()

    def serialize(self) -> bytes:
        data = [0, self.pubkey, self.created_at, self.kind, self.tags, self.content]
        data_str = json.dumps(data, separators=(',', ':'), ensure_ascii=False)
        return data_str.encode()

    def compute_id(self):
        self.id = sha256(self.serialize()).hexdigest()

    @classmethod
    def from_dict(cls, msg: dict) -> 'Event':
        # "id" is ignore, as it will be computed from the contents
        return Event(
            content=msg['content'],
            pubkey=msg['pubkey'],
            created_at=msg['created_at'],
            kind=msg['kind'],
            tags=msg['tags'],
            sig=msg['sig'],
        )

    def add_pubkey_ref(self, pubkey: str):
        """Adds a reference to a pubkey as a 'p' tag."""
        self.tags.append(['p', pubkey])
        self.compute_id()

    def has_pubkey_ref(self, pubkey: str):
        for tag_type, tag in self.tags:
            if tag_type == 'p' and tag == pubkey:
                return True
        return False

    def add_event_ref(self, event_id: str):
        """Adds a reference to an event_id as an 'e' tag."""
        self.tags.append(['e', event_id])
        self.compute_id()

    def has_event_ref(self, event_id: str):
        """Check if a e tag to the given event_id exists."""
        for tag_type, tag in self.tags:
            if tag_type == 'e' and tag == event_id:
                return True
        return False

    def get_tag_dict(self):
        """Returns all tags as dict."""
        ret = {}
        tag_types = self.get_tag_types()
        for t in tag_types:
            ret[t] = self.get_tag_list(tag_type=t)
        return ret

    def get_tag_list(self, tag_type: str = 'e'):
        """Returns all tags of given type as list."""
        ret = []
        for _tag_type, _tag in self.tags:
            if _tag_type == tag_type:
                ret.append(_tag)
        return ret

    def get_tag_types(self):
        """Returns list of all included tag types."""
        ret = []
        for tag_type, _ in self.tags:
            if tag_type not in ret:
                ret.append(tag_type)
        return ret

    def get_tag_count(self, tag_type: str = 'e'):
        """Returns all tags of given type as list."""
        count = 0
        for _tag_type, _tag in self.tags:
            if _tag_type == tag_type:
                count += 1
        return count

    def encrypt_dm(
        self, private_key_hex: str, cleartext_content: str, recipient_pubkey: str
    ) -> None:
        if self.kind != EventKind.ENCRYPTED_DIRECT_MESSAGE:
            raise Exception("Wrong event kind, needs to be ENCRYPTED_DIRECT_MESSAGE")
        if not self.has_pubkey_ref(recipient_pubkey):
            # Must specify the DM recipient's pubkey in a 'p' tag
            self.add_pubkey_ref(recipient_pubkey)
        sk = PrivateKey(bytes.fromhex(private_key_hex))
        encrypted_message = sk.encrypt_message(
            message=cleartext_content, public_key_hex=recipient_pubkey
        )
        self.content = encrypted_message

    def bech32(self) -> str:
        """bech32-encoded entities (N-19)

        :return: note id as bech32 encoding with note prefix
        """
        self.compute_id()
        return bech32_encode(binascii.unhexlify(self.id), "note")

    def sign(self, private_key_hex: str) -> None:
        """signs the event with the private key and stored the signature in self.sig.
        The pubkey from the event is replaced and the note id recomputed.

        :param private_key_hex: private key as hex string
        """
        if self.kind == EventKind.ENCRYPTED_DIRECT_MESSAGE and self.content is None:
            raise Exception("Message is not yet encrypted!")
        sk = PrivateKey(bytes.fromhex(private_key_hex))
        self.pubkey = sk.public_key.hex()
        self.compute_id()
        sig = sk.sign(bytes.fromhex(self.id))
        self.sig = sig.hex()

    def verify(self) -> bool:
        pub_key = PublicKey.from_hex(self.pubkey)
        self.compute_id()
        return pub_key.verify(bytes.fromhex(self.sig), bytes.fromhex(self.id))

    def date_time(self):
        return datetime.datetime.utcfromtimestamp(self.created_at)

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "pubkey": self.pubkey,
            "created_at": self.created_at,
            "kind": self.kind,
            "tags": self.tags,
            "content": self.content,
            "sig": self.sig,
        }

    def to_message(self) -> str:
        return json.dumps(
            [
                ClientMessageType.EVENT,
                self.to_dict(),
            ]
        )

    def __repr__(self):
        note_id = self.bech32()
        return f'Event({note_id[:10]}...{note_id[-10:]})'

    def __str__(self):
        return self.to_message()



import json
from typing import Union

from tornado import gen
from tornado.ioloop import IOLoop
from tornado.websocket import websocket_connect

from .base_relay import BaseRelay, RelayPolicy, RelayProxyConnectionConfig
from .message_pool import MessagePool
from .message_type import RelayMessageType


class TornadoRelay(BaseRelay):
    def __init__(
        self,
        url: str,
        message_pool: MessagePool,
        io_loop: IOLoop,
        policy: RelayPolicy = RelayPolicy(),
        ssl_options: dict = None,
        proxy_config: Union[None, RelayProxyConnectionConfig] = None,
    ) -> None:
        super().__init__(url, message_pool, policy, ssl_options, proxy_config)
        self.io_loop = io_loop

    @property
    def is_connected(self) -> bool:
        return self.ws is not None

    @gen.coroutine
    def connect(self):
        yield websocket_connect(
            self.url,
            callback=self.maybe_retry_connection,
            on_message_callback=self.on_message,
            ping_interval=60,
            ping_timeout=120,
        )

    def maybe_retry_connection(self, future) -> None:
        try:
            self.ws = future.result()
            self.publish(self.request)
        except Exception:
            print("Could not reconnect, retrying in 3 seconds...")
            self.io_loop.call_later(3, self.connect)

    def on_message(self, message):
        if message is None:
            self.connect()
        else:
            if self._is_valid_message(message):
                message_json = json.loads(message)
                message_type = message_json[0]
                if message_type == RelayMessageType.EVENT:
                    # event = Event.from_dict(message_json[2])
                    # print(event.to_message())
                    self.message_pool.add_message(message, self.url)
                elif message_type == RelayMessageType.END_OF_STORED_EVENTS:
                    self.close()

    def start(self):
        self.connect()

    def close(self):
        if self.ws is not None:
            self.ws.close()
            self.io_loop.stop()

    def publish(self, message: str):
        # print(message)
        self.ws.write_message(message)



import base64
import binascii
import secrets
from hashlib import sha256
from typing import Optional

import coincurve as secp256k1
from coincurve._libsecp256k1 import ffi, lib
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from .bech32 import bech32_decode, bech32_encode
from .delegation import Delegation

HAS_ECDH = hasattr(lib, 'secp256k1_ecdh')


class PublicKey:
    def __init__(self, raw_bytes: bytes) -> None:
        """
        :param raw_bytes: The formatted public key.
        :type data: bytes
        """
        if isinstance(raw_bytes, PrivateKey):
            self.raw_bytes = raw_bytes.public_key.raw_bytes
        elif isinstance(raw_bytes, secp256k1.keys.PublicKey):
            self.raw_bytes = raw_bytes.format(compressed=True)[2:]
        elif isinstance(raw_bytes, secp256k1.keys.PublicKeyXOnly):
            self.raw_bytes = raw_bytes.format()
        elif isinstance(raw_bytes, str):
            self.raw_bytes = binascii.unhexlify(raw_bytes)
        else:
            self.raw_bytes = raw_bytes

    def bech32(self) -> str:
        return bech32_encode(self.raw_bytes, "npub")

    @property
    def npub(self):
        return self.bech32()

    def hex(self) -> str:
        return self.raw_bytes.hex()

    def verify(self, sig: bytes, message: bytes) -> bool:
        pk = secp256k1.PublicKeyXOnly(self.raw_bytes)
        return pk.verify(sig, message)

    @classmethod
    def from_hex(cls, hex: str) -> 'PublicKey':
        return cls(bytes.fromhex(hex))

    @classmethod
    def from_npub(cls, npub: str):
        """Load a PublicKey from its bech32/npub form."""
        return cls(bytes(bech32_decode(npub)))

    def __repr__(self):
        pubkey = self.bech32()
        return f'PublicKey({pubkey[:10]}...{pubkey[-10:]})'

    def __str__(self):
        """Return public key in hex form
        :return: string
        :rtype: str
        """
        return self.hex()

    def __bytes__(self):
        """Return raw bytes
        :return: Raw bytes
        :rtype: bytes
        """
        return self.raw_bytes


class PrivateKey:
    def __init__(self, raw_secret: Optional[bytes] = None) -> None:
        """
        :param raw_secret: The secret used to initialize the private key.
                           If not provided or `None`, a new key will be generated.
        :type raw_secret: bytes
        """
        if raw_secret is not None:
            self.raw_secret = raw_secret
        else:
            self.raw_secret = secrets.token_bytes(32)

        sk = secp256k1.PrivateKey(self.raw_secret)
        self.public_key = PublicKey(sk.public_key_xonly)

    @classmethod
    def from_nsec(cls, nsec: str):
        """Load a PrivateKey from its bech32/nsec form.

        :param nsec: the nsec key to be imported
        """
        return cls(bytes(bech32_decode(nsec)))

    @classmethod
    def from_hex(cls, hex: str):
        """Load a PrivateKey from its hex form."""
        return cls(binascii.unhexlify(hex))

    def bech32(self) -> str:
        return bech32_encode(self.raw_secret, "nsec")

    @property
    def nsec(self):
        return self.bech32()

    def hex(self) -> str:
        return self.raw_secret.hex()

    def ecdh(self, public_key_hex):
        assert public_key_hex, "No public key defined"
        if not HAS_ECDH:
            raise Exception("secp256k1_ecdh not enabled")
        sk = secp256k1.PrivateKey(self.raw_secret)
        result = ffi.new('char [32]')
        pk = secp256k1.PublicKey(bytes.fromhex("02" + public_key_hex))
        res = lib.secp256k1_ecdh(
            sk.context.ctx, result, pk.public_key, self.raw_secret, copy_x, ffi.NULL
        )
        if not res:
            raise Exception(f'invalid scalar ({res})')

        return bytes(ffi.buffer(result, 32))

    def tweak_add(self, scalar: bytes) -> bytes:
        sk = secp256k1.PrivateKey(self.raw_secret)
        return sk.add(scalar)

    def compute_shared_secret(self, public_key_hex: str) -> bytes:
        return self.ecdh(public_key_hex)

    def encrypt_message(self, message: str, public_key_hex: str) -> str:
        padder = padding.PKCS7(128).padder()
        padded_data = padder.update(message.encode()) + padder.finalize()

        iv = secrets.token_bytes(16)
        cipher = Cipher(
            algorithms.AES(self.compute_shared_secret(public_key_hex)), modes.CBC(iv)
        )

        encryptor = cipher.encryptor()
        encrypted_message = encryptor.update(padded_data) + encryptor.finalize()

        ret_part1 = base64.b64encode(encrypted_message).decode()
        ret_part2 = base64.b64encode(iv).decode()
        return f"{ret_part1}?iv={ret_part2}"

    def decrypt_message(self, encoded_message: str, public_key_hex: str) -> str:
        encoded_data = encoded_message.split('?iv=')
        encoded_content, encoded_iv = encoded_data[0], encoded_data[1]

        iv = base64.b64decode(encoded_iv)
        cipher = Cipher(
            algorithms.AES(self.compute_shared_secret(public_key_hex)), modes.CBC(iv)
        )
        encrypted_content = base64.b64decode(encoded_content)

        decryptor = cipher.decryptor()
        decrypted_message = decryptor.update(encrypted_content) + decryptor.finalize()

        unpadder = padding.PKCS7(128).unpadder()
        unpadded_data = unpadder.update(decrypted_message) + unpadder.finalize()

        return unpadded_data.decode()

    def sign(self, message: bytes, aux_randomness: bytes = b'') -> str:
        sk = secp256k1.PrivateKey(self.raw_secret)
        return sk.sign_schnorr(message, aux_randomness)

    def sign_delegation(self, delegation: Delegation) -> None:
        delegation.signature = self.sign(
            sha256(delegation.delegation_token.encode()).digest()
        )

    def __eq__(self, other):
        return self.raw_secret == other.raw_secret

    def __repr__(self):
        pubkey = self.public_key.bech32()
        return f'PrivateKey({pubkey[:10]}...{pubkey[-10:]})'

    def __str__(self):
        """Return private key in hex form
        :return: hex string
        :rtype: str
        """
        return self.hex()

    def __bytes__(self):
        """Return raw bytes
        :return: Raw bytes
        :rtype: bytes
        """
        return self.raw_secret


@ffi.callback(
    "int (unsigned char *, const unsigned char *, const unsigned char *, void *)"
)
def copy_x(output, x32, y32, data):
    ffi.memmove(output, x32, 32)
    return 1



from . import bech32_util


def bech32_decode(bech32_str: str):
    """Loads bytes from its bech32/nsec form."""
    hrp, data, spec = bech32_util.bech32_decode(bech32_str)
    raw_secret = bech32_util.convertbits(data, 5, 8)[:-1]
    return bytes(raw_secret)


def bech32_encode(raw_bytes: bytes, prefix: str) -> str:
    converted_bits = bech32_util.convertbits(raw_bytes, 8, 5)
    return bech32_util.bech32_encode(
        prefix, converted_bits, bech32_util.Encoding.BECH32
    )



import json
from dataclasses import dataclass
from threading import Lock
from typing import Dict

from .event import Event
from .filters import FiltersList
from .message_pool import MessagePool
from .message_type import RelayMessageType
from .request import Request
from .subscription import Subscription


@dataclass
class RelayPolicy:
    should_read: bool = True
    should_write: bool = True

    def to_json_object(self) -> Dict[str, bool]:
        return {"read": self.should_read, "write": self.should_write}


@dataclass
class RelayProxyConnectionConfig:
    host: str
    port: int
    type: str


class BaseRelay:
    def __init__(
        self,
        url: str,
        message_pool: MessagePool,
        policy: RelayPolicy = RelayPolicy(),
        ssl_options: dict = None,
        proxy_config: RelayProxyConnectionConfig = None,
    ) -> None:
        self.url = url
        self.message_pool = message_pool
        self.policy = policy
        self.ssl_options = ssl_options
        self.proxy_config = proxy_config
        self.lock: Lock = Lock()
        self.ws = None
        self.subscriptions: dict[str, Subscription] = {}
        self.connected: bool = False
        self.reconnect: bool = True
        self.error_counter: int = 0
        self.error_threshold: int = 10
        self.request: str = ""

    def __repr__(self):
        return json.dumps(self.to_json_object(), indent=2)

    def to_json_object(self) -> dict:
        return {
            "url": self.url,
            "policy": self.policy.to_json_object(),
            "subscriptions": [
                subscription.to_json_object()
                for subscription in self.subscriptions.values()
            ],
            "request": self.request,
        }

    def add_subscription(self, id, filters: FiltersList):
        with self.lock:
            self.subscriptions[id] = Subscription(id, filters)
            self.request = Request(id, filters).to_message()

    def close_subscription(self, id: str) -> None:
        with self.lock:
            self.subscriptions.pop(id, None)

    def update_subscription(self, id: str, filters: FiltersList) -> None:
        with self.lock:
            subscription = self.subscriptions[id]
            subscription.filters = filters
            self.request = Request(id, filters).to_message()

    def _is_valid_message(self, message: str) -> bool:
        message = message.strip("\n")
        if not message or message[0] != '[' or message[-1] != ']':
            return False

        message_json = json.loads(message)
        message_type = message_json[0]
        if not RelayMessageType.is_valid(message_type):
            return False
        if message_type == RelayMessageType.EVENT:
            if not len(message_json) == 3:
                return False

            subscription_id = message_json[1]
            with self.lock:
                if subscription_id not in self.subscriptions:
                    return False

            event = Event.from_dict(message_json[2])
            if not event.verify():
                return False

            with self.lock:
                subscription = self.subscriptions[subscription_id]

            if subscription.filtersList and not subscription.filtersList.match(event):
                return False

        return True



import unittest

from typer.testing import CliRunner

from pynostr.cli import app


class TestEvent(unittest.TestCase):
    def test_keygen(self):
        runner = CliRunner()
        result = runner.invoke(app, ['keygen'])
        self.assertEqual(result.exit_code, 0)

    def test_convert(self):
        runner = CliRunner()
        result = runner.invoke(
            app,
            [
                'convert',
                'npub1mg2nzunrsk9df94zr3uudhzltnu6lzq2muax09xmhu5gxxrvnkqsvpjg3p',
            ],
        )
        self.assertEqual(result.exit_code, 0)



import json
from dataclasses import dataclass

from .filters import FiltersList
from .message_type import ClientMessageType


@dataclass
class Request:
    subscription_id: str
    filtersList: FiltersList

    def to_message(self) -> str:
        message = [ClientMessageType.REQUEST, self.subscription_id]
        message.extend(self.filtersList.to_json_array())
        return json.dumps(message)



"""Forked from https://github.com/jeffthibault/python-nostr.git."""
import time
import unittest

from pynostr.event import Event, EventKind
from pynostr.key import PrivateKey


class TestEvent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.sender_pk = PrivateKey()
        cls.sender_pubkey = cls.sender_pk.public_key.hex()

    def test_note_event(self):
        pk1 = PrivateKey.from_hex(
            "964b29795d621cdacf05fd94fb23206c88742db1fa50b34d7545f3a2221d8124"
        )
        event = Event("Hello Nostr!", pk1.public_key.hex())
        event.created_at = 1671406583
        event.compute_id()
        event.sign(pk1.hex())
        self.assertEqual(
            "23411895658d374ec922adf774a70172290b2c738ae67815bd8945e5d8fff3bb", event.id
        )
        self.assertTrue(event.verify())

    def test_event_default_time(self):
        public_key = PrivateKey().public_key.hex()
        event1 = Event(pubkey=public_key, content='test event')
        time.sleep(1.5)
        event2 = Event(pubkey=public_key, content='test event')
        self.assertTrue(event1.created_at < event2.created_at)

    def test_add_event_ref(self):
        some_event_id = "some_event_id"
        event = Event(content="Adding an 'e' tag")
        event.add_event_ref(some_event_id)
        self.assertTrue(['e', some_event_id] in event.tags)
        self.assertEqual(event.get_tag_count('e'), 1)
        self.assertEqual(event.get_tag_count('p'), 0)
        self.assertEqual(event.get_tag_list('e'), [some_event_id])
        self.assertEqual(event.get_tag_list('p'), [])
        self.assertEqual(event.get_tag_types(), ['e'])
        self.assertEqual(event.get_tag_dict(), {"e": [some_event_id]})
        self.assertTrue(event.has_event_ref(some_event_id))

    def test_add_pubkey_ref(self):
        some_pubkey = "some_pubkey"
        event = Event(content="Adding a 'p' tag")
        event.add_pubkey_ref(some_pubkey)
        self.assertTrue(['p', some_pubkey] in event.tags)
        self.assertEqual(event.get_tag_count('p'), 1)
        self.assertEqual(event.get_tag_count('e'), 0)
        self.assertEqual(event.get_tag_list('p'), [some_pubkey])
        self.assertEqual(event.get_tag_list('e'), [])
        self.assertEqual(event.get_tag_types(), ['p'])
        self.assertEqual(event.get_tag_dict(), {"p": [some_pubkey]})
        self.assertTrue(event.has_pubkey_ref(some_pubkey))

    def test_sign_event_is_valid(self):
        """Sign should create a signature that can be verified against Event.id."""
        event = Event(content="Hello, world!")
        event.sign(self.sender_pk.hex())
        self.assertTrue(event.verify())

    def test_sign_event_adds_pubkey(self):
        """Sign should add the sender's pubkey if not already specified."""
        event = Event(content="Hello, world!")

        # The event's public_key hasn't been specified yet
        self.assertTrue(event.pubkey is None)
        event.sign(self.sender_pk.hex())

        # PrivateKey.sign() should have populated public_key
        self.assertEqual(event.pubkey, self.sender_pubkey)

    def test_dict_roundtrip(self):
        """Conversion to dict and back result in same object."""
        event = Event(
            content='test event',
            created_at=12345678,
            kind=1,
        )
        event.add_pubkey_ref("some_pubkey")

        got = Event.from_dict(event.to_dict())
        self.assertEqual(got, event)


class TestEncryptedDirectMessage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.sender_pk = PrivateKey.from_hex(
            "29307c4354b7d9d311d2cec4878c0de56c93a921d300273c19577e9004de3c9f"
        )
        cls.sender_pubkey = cls.sender_pk.public_key.hex()
        cls.recipient_pk = PrivateKey.from_hex(
            "4138d1b6dde34f81c38cef2630429e85847dd5b70508e37f53c844f66f19f983"
        )
        cls.recipient_pubkey = cls.recipient_pk.public_key.hex()

    def test_content_field_not_allowed(self):
        """Should not let users instantiate a new DM with `content` field data."""
        with self.assertRaisesRegex(Exception, "cannot use"):
            Event(content="My message!", kind=EventKind.ENCRYPTED_DIRECT_MESSAGE)

    def test_recipient_p_tag(self):
        """Should generate recipient 'p' tag."""
        dm = Event(kind=EventKind.ENCRYPTED_DIRECT_MESSAGE)
        dm.encrypt_dm(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content="Secret message!",
        )
        assert ['p', self.recipient_pubkey] in dm.tags

    def test_encrypt_dm(self):
        """Should encrypt a DM and populate its `content` field with ciphertext that
        either party can decrypt."""
        message1 = "Test"
        message2 = "Test2"
        message3 = "Test3"
        message4 = "Test4"

        dm1 = Event(kind=EventKind.ENCRYPTED_DIRECT_MESSAGE)
        self.assertTrue(dm1.content is None)
        dm1.encrypt_dm(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content=message1,
        )
        dm2 = Event(kind=EventKind.ENCRYPTED_DIRECT_MESSAGE)
        dm2.encrypt_dm(
            self.recipient_pk.hex(),
            recipient_pubkey=self.sender_pubkey,
            cleartext_content=message2,
        )
        dm3 = Event(kind=EventKind.ENCRYPTED_DIRECT_MESSAGE)
        dm3.encrypt_dm(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content=message3,
        )
        dm4 = Event(kind=EventKind.ENCRYPTED_DIRECT_MESSAGE)
        dm4.encrypt_dm(
            self.recipient_pk.hex(),
            recipient_pubkey=self.sender_pubkey,
            cleartext_content=message4,
        )

        # After encrypting, the content field should now be populated
        self.assertTrue(dm1.content is not None)

        # Sender should be able to decrypt
        decm1_sender = self.sender_pk.decrypt_message(
            encoded_message=dm1.content, public_key_hex=self.recipient_pubkey
        )
        decm2_sender = self.sender_pk.decrypt_message(
            encoded_message=dm2.content, public_key_hex=self.recipient_pubkey
        )
        decm3_sender = self.sender_pk.decrypt_message(
            encoded_message=dm3.content, public_key_hex=self.recipient_pubkey
        )
        decm4_sender = self.sender_pk.decrypt_message(
            encoded_message=dm4.content, public_key_hex=self.recipient_pubkey
        )

        decm1_receiver = self.recipient_pk.decrypt_message(
            encoded_message=dm1.content, public_key_hex=self.sender_pubkey
        )
        decm2_receiver = self.recipient_pk.decrypt_message(
            encoded_message=dm2.content, public_key_hex=self.sender_pubkey
        )
        decm3_receiver = self.recipient_pk.decrypt_message(
            encoded_message=dm3.content, public_key_hex=self.sender_pubkey
        )
        decm4_receiver = self.recipient_pk.decrypt_message(
            encoded_message=dm4.content, public_key_hex=self.sender_pubkey
        )
        self.assertEqual(message1, decm1_receiver)
        self.assertEqual(message1, decm1_sender)
        self.assertEqual(message2, decm2_receiver)
        self.assertEqual(message2, decm2_sender)
        self.assertEqual(message3, decm3_receiver)
        self.assertEqual(message3, decm3_sender)
        self.assertEqual(message4, decm4_receiver)
        self.assertEqual(message4, decm4_sender)

    def test_decrypt_dm(self):
        message1 = "Test"
        message2 = "Test2"
        message3 = "Test3"
        message4 = "Test4"
        encrypt1 = "VOqWLiW4wv8+fDsNC00a1w==?iv=LSIH1sk13Mw09PV8Z80sag=="
        encrypt2 = "abZBRLPta8888xDkg6pUWA==?iv=Gj5KOUbFqREhSdbMENRKEg=="
        encrypt3 = "w2AsXNN0EysjG6/E/GZWPg==?iv=3c7qsPxSOckGeqjjpwQQ4A=="
        encrypt4 = "nBfP5P2GEEOlfNYMoxADDg==?iv=VsFd7eE8BfoyDJpfQ7kjhQ=="

        dm_s1 = self.sender_pk.decrypt_message(
            encoded_message=encrypt1, public_key_hex=self.recipient_pubkey
        )
        dm_s2 = self.sender_pk.decrypt_message(
            encoded_message=encrypt2, public_key_hex=self.recipient_pubkey
        )
        dm_s3 = self.sender_pk.decrypt_message(
            encoded_message=encrypt3, public_key_hex=self.recipient_pubkey
        )
        dm_s4 = self.sender_pk.decrypt_message(
            encoded_message=encrypt4, public_key_hex=self.recipient_pubkey
        )
        self.assertEqual(dm_s1, message1)
        self.assertEqual(dm_s2, message2)
        self.assertEqual(dm_s3, message3)
        self.assertEqual(dm_s4, message4)

        dm_r1 = self.recipient_pk.decrypt_message(
            encoded_message=encrypt1, public_key_hex=self.sender_pubkey
        )
        dm_r2 = self.recipient_pk.decrypt_message(
            encoded_message=encrypt2, public_key_hex=self.sender_pubkey
        )
        dm_r3 = self.recipient_pk.decrypt_message(
            encoded_message=encrypt3, public_key_hex=self.sender_pubkey
        )
        dm_r4 = self.recipient_pk.decrypt_message(
            encoded_message=encrypt4, public_key_hex=self.sender_pubkey
        )
        self.assertEqual(dm_r1, message1)
        self.assertEqual(dm_r2, message2)
        self.assertEqual(dm_r3, message3)
        self.assertEqual(dm_r4, message4)

    def test_sign_encrypts_dm(self):
        """`sign` should encrypt a DM that hasn't been encrypted yet."""
        dm = Event(kind=EventKind.ENCRYPTED_DIRECT_MESSAGE)
        self.assertTrue(dm.content is None)
        dm.encrypt_dm(
            self.sender_pk.hex(),
            recipient_pubkey=self.recipient_pubkey,
            cleartext_content="Some DM message",
        )

        dm.sign(self.sender_pk.hex())

        self.assertTrue(dm.content is not None)

    def test_shared_secret(self):
        sender_pk = "29307c4354b7d9d311d2cec4878c0de56c93a921d300273c19577e9004de3c9f"
        recipient_pk = (
            "4138d1b6dde34f81c38cef2630429e85847dd5b70508e37f53c844f66f19f983"
        )

        sender_pubkey = PrivateKey.from_hex(sender_pk).public_key.hex()
        recipient_pubkey = PrivateKey.from_hex(recipient_pk).public_key.hex()

        shared_secret1 = PrivateKey.from_hex(sender_pk).compute_shared_secret(
            recipient_pubkey
        )
        shared_secret2 = PrivateKey.from_hex(recipient_pk).compute_shared_secret(
            sender_pubkey
        )
        self.assertEqual(shared_secret1, shared_secret2)

    def test_decrypt_event(self):
        dm1 = Event.from_dict(
            {
                "id": "46c76ec67d03babdb840254b3667585143cc499497b0a6a40aedc9ce2de416"
                "70",
                "pubkey": "f3c25355c29f64ea8e9b4e11b583ac0a7d0d8235f156cffec2b73e5756a"
                "ab206",
                "created_at": 1674819397,
                "kind": 4,
                "tags": [
                    [
                        "p",
                        "a1db8e8b047e1350958a55e0a853151d0e1f685fa5cf3772e01bccc5aa5c"
                        "b2eb",
                    ]
                ],
                "content": "VOqWLiW4wv8+fDsNC00a1w==?iv=LSIH1sk13Mw09PV8Z80sag==",
                "sig": "982e825d32e78bf695a2511e14d2958e0d480bfa338e12307e5eb7ecad3b"
                "d86d53c2d508df9083a0fc25b44c44ae4c622f7405bbc2ca723c5dbc98da48214f16",
            }
        )
        self.assertTrue(dm1.verify())



from dataclasses import dataclass

from .filters import FiltersList


@dataclass
class Subscription:
    id: str
    filtersList: FiltersList

    def __post_init__(self):
        if not isinstance(self.id, str):
            raise TypeError("Argument 'id' must be of type str")

    def to_json_object(self):
        return {"id": self.id, "filters": self.filtersList.to_json_array()}



import binascii
import unittest

import tlv8

from pynostr.bech32 import bech32_decode, bech32_encode


class TestBech32(unittest.TestCase):
    def test_npub(self):
        npub = "npub180cvv07tjdrrgpa0j7j7tmnyl2yr6yr7l8j4s3evf6u64th6gkwsyjh6w6"
        pub_key = "3bf0c63fcb93463407af97a5e5ee64fa883d107ef9e558472c4eb9aaaefa459d"
        self.assertEqual(bytes(bech32_decode(npub)).hex(), pub_key)
        self.assertEqual(bech32_encode(binascii.unhexlify(pub_key), "npub"), npub)

    def test_nsec(self):
        npub = "nsec180cvv07tjdrrgpa0j7j7tmnyl2yr6yr7l8j4s3evf6u64th6gkwsgyumg0"
        pub_key = "3bf0c63fcb93463407af97a5e5ee64fa883d107ef9e558472c4eb9aaaefa459d"
        self.assertEqual(bytes(bech32_decode(npub)).hex(), pub_key)
        self.assertEqual(bech32_encode(binascii.unhexlify(pub_key), "nsec"), npub)

    def test_nprofile(self):
        nprofile = (
            "nprofile1qqsrhuxx8l9ex335q7he0f09aej04zpazpl0ne2cgukyaw"
            + "d24mayt8gpp4mhxue69uhhytnc9e3k7mgpz4mhxue69uhkg6nzv9e"
            + "juumpv34kytnrdaksjlyr9p"
        )
        pub_key = "3bf0c63fcb93463407af97a5e5ee64fa883d107ef9e558472c4eb9aaaefa459d"
        relay1 = "wss://r.x.com"
        relay2 = "wss://djbas.sadkb.com"
        decode = bytes(bech32_decode(nprofile))

        data = tlv8.deep_decode(decode)
        self.assertEqual(data[0].type_id, 0)
        self.assertEqual(binascii.hexlify(data[0].data).decode(), pub_key)

        self.assertEqual(data[1].type_id, 1)
        self.assertEqual(data[1].data.decode(), relay1)

        self.assertEqual(data[2].type_id, 1)
        self.assertEqual(data[2].data.decode(), relay2)

        structure = [
            tlv8.Entry(0, binascii.unhexlify(pub_key)),
            tlv8.Entry(1, relay1.encode()),
            tlv8.Entry(1, relay2.encode()),
        ]
        bytes_data = tlv8.encode(structure)

        bytes_data = b''
        for entry in structure:
            bytes_data += entry.encode()

        self.assertEqual(bech32_encode(bytes_data, "nprofile"), nprofile)



import logging

import click
import typer

from pynostr.key import PrivateKey, PublicKey

log = logging.getLogger(__name__)
app = typer.Typer()

state = {"verbose": 3}


@app.command()
def keygen():
    """Creates a private and public key."""
    private_key = PrivateKey()
    public_key = private_key.public_key
    click.echo(f"Private key: {private_key.bech32()}")
    click.echo(f"Public key: {public_key.bech32()}")


@app.command()
def convert(npub: str):
    """Converts npub key to hex."""
    public_key = PublicKey.from_npub(npub)
    click.echo(f"npub: {public_key.bech32()}")
    click.echo(f"hex: {public_key.hex()}")


@app.callback()
def main(verbose: int = 3):
    """Python CLI for nostr, enjoy."""
    # Logging
    state["verbose"] = verbose
    log = logging.getLogger(__name__)
    verbosity = ["critical", "error", "warn", "info", "debug"][int(min(verbose, 4))]
    log.setLevel(getattr(logging, verbosity.upper()))
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    ch = logging.StreamHandler()
    ch.setLevel(getattr(logging, verbosity.upper()))
    ch.setFormatter(formatter)
    log.addHandler(ch)


if __name__ == "__main__":
    app()



"""Forked from https://github.com/jeffthibault/python-nostr.git."""
import threading
import time
from typing import Union

from websocket import WebSocketApp

from .base_relay import BaseRelay, RelayPolicy, RelayProxyConnectionConfig
from .message_pool import MessagePool


class Relay(BaseRelay):
    def __init__(
        self,
        url: str,
        message_pool: MessagePool,
        policy: RelayPolicy = RelayPolicy(),
        ssl_options: dict = None,
        proxy_config: Union[None, RelayProxyConnectionConfig] = None,
    ) -> None:
        super().__init__(url, message_pool, policy, ssl_options, proxy_config)
        self.ws: WebSocketApp = WebSocketApp(
            self.url,
            on_open=self._on_open,
            on_message=self._on_message,
            on_error=self._on_error,
            on_close=self._on_close,
        )

    @property
    def is_connected(self) -> bool:
        return False if self.ws.sock is None else self.ws.sock.connected

    def open_connections(self, ssl_options: dict = None):
        if ssl_options is None:
            ssl_options = {}
        self.ssl_options = ssl_options
        if not self.is_connected:
            threading.Thread(target=self.connect, name=f"{self.url}-thread").start()
        time.sleep(2)
        assert self.is_connected
        self._is_connected = True

    def close(self):
        if self.ws.sock is not None:
            self.ws.close()

    def close_connections(self):
        self.close()

    def connect(self):
        self.ws.run_forever(
            sslopt=self.ssl_options,
            http_proxy_host=None
            if self.proxy_config is None
            else self.proxy_config.host,
            http_proxy_port=None
            if self.proxy_config is None
            else self.proxy_config.port,
            proxy_type=None if self.proxy_config is None else self.proxy_config.type,
        )

    def check_reconnect(self):
        try:
            self.close()
        except Exception:
            pass
        self.connected = False
        if self.reconnect:
            time.sleep(1)
            self.connect()

    def publish(self, message: str):
        if self.connected:
            self.ws.send(message)

    def _on_open(self, class_obj):
        self.connected = True

    def _on_close(self, class_obj, status_code, message):
        self.connected = False

    def _on_message(self, class_obj, message: str):
        if message is None:
            print("Empty message received")
        else:
            if self._is_valid_message(message):
                self.message_pool.add_message(message, self.url)

    def _on_error(self, class_obj, error):
        self.connected = False
        self.error_counter += 1
        if self.error_threshold and self.error_counter > self.error_threshold:
            pass
        else:
            self.check_reconnect()
